@@ -1,13 +1,24 @@
-import os
 from ast import literal_eval
 from getpass import getpass
-
 import io
+import os
 try:
-    from json import loads, JSONDecodeError
+    from pathlib import Path
+except ImportError:
+    # do not care: only used for type hinting
+    pass
+import sys
+
+if sys.version >= "3":
+    from io import StringIO
+else:
+    from StringIO import StringIO
+
+try:
+    from json import loads, JSONDecodeError, dumps
 except ImportError:
     # python 2
-    from json import loads
+    from json import loads, dumps
     JSONDecodeError = ValueError
 try:
     FileNotFoundError
@@ -28,11 +39,10 @@ except ImportError:
 
 try:
     # noinspection PyUnresolvedReferences
-    from typing import Dict
+    from typing import Dict, Union
 except ImportError:
     pass
 
-
 ODS_BASE_URL_TEMPLATE = "https://%s.opendatasoft.com"
 ENV_ODS_APIKEY = 'ODS_APIKEY'
 KR_DEFAULT_USERNAME = 'apikey_user'
@@ -44,8 +54,9 @@ class ODSClient(object):
     `https://<platform_id>.opendatasoft.com` with `platform_id='public'`. One can change either customize the platform
     id through the `platform_id` constructor argument, or the whole base url with `base_url`.
 
-    A client instance offers methods to interact with the various ODS API. Currently two high-level methods are
-    provided: `<client>.get_whole_dataset(dataset_id, ...)` and `<client>.get_whole_dataframe(dataset_id, ...)`
+    A client instance offers methods to interact with the various ODS API. Currently three high-level methods are
+    provided: `<client>.get_whole_dataset(dataset_id, ...)`, `<client>.get_whole_dataframe(dataset_id, ...)`
+    and `<client>.push_dataset_realtime(dataset_id, ...)`.
 
     You can customize the `requests.Session` object used for the HTTPS transport using `requests_session`.
 
@@ -72,12 +83,13 @@ class ODSClient(object):
     is the one you think you have configured through one of the above methods.
 
     """
+
     def __init__(self,
                  platform_id='public',                          # type: str
                  base_url=None,                                 # type: str
                  enforce_apikey=False,                          # type: bool
                  apikey=None,                                   # type: str
-                 apikey_filepath='ods.apikey',                  # type: str
+                 apikey_filepath='ods.apikey',                  # type: Union[str, Path]
                  use_keyring=True,                              # type: bool
                  keyring_entries_username=KR_DEFAULT_USERNAME,  # type: str
                  requests_session=None                          # type: Session
@@ -110,7 +122,7 @@ class ODSClient(object):
             if platform_id != 'public' and platform_id is not None:
                 raise ValueError("Only one of `platform_id` and `base_url` should be provided. Received "
                                  "platform_id='%s' and base_url='%s'" % (platform_id, base_url))
-            # remove trailing slashs
+            # remove trailing slashes
             while base_url.endswith('/'):
                 base_url = base_url[:-1]
             self.base_url = base_url
@@ -130,7 +142,7 @@ class ODSClient(object):
         elif apikey_filepath is not None:
             try:
                 # read the api key from the file
-                with open(apikey_filepath) as f:
+                with open(str(apikey_filepath)) as f:
                     self.apikey = f.read()
             except FileNotFoundError:
                 self.apikey = None
@@ -153,6 +165,8 @@ class ODSClient(object):
     def get_whole_dataframe(self,
                             dataset_id,                  # type: str
                             use_labels_for_header=True,  # type: bool
+                            tqdm=False,                  # type: bool
+                            block_size=1024,             # type: int
                             **other_opts
                             ):
         """
@@ -160,6 +174,8 @@ class ODSClient(object):
 
         :param dataset_id:
         :param use_labels_for_header:
+        :param tqdm: a boolean indicating if a progress bar using tqdm should be displayed. tqdm should be installed
+        :param block_size: an int block size used in streaming mode when tqdm is used
         :param other_opts:
         :return:
         """
@@ -192,7 +208,21 @@ class ODSClient(object):
         # Execute call in stream mode
         result = self._http_call(url, params=opts, stream=True)
         # print(iterable_to_stream(result.iter_content()).read())
-        df = pd.read_csv(iterable_to_stream(result.iter_content()), sep=';')
+        # noinspection PyTypeChecker
+
+        if tqdm:
+            from tqdm import tqdm as _tqdm
+
+            total_size = int(result.headers.get('Content-Length', 0))
+            with _tqdm(desc=url, total=total_size,
+                       unit='B' if block_size == 1024 else 'it',
+                       unit_scale=True,
+                       unit_divisor=block_size
+                       ) as bar:
+                df = pd.read_csv(iterable_to_stream(result.iter_content(), buffer_size=block_size, progressbar=bar),
+                                 sep=';')
+        else:
+            df = pd.read_csv(iterable_to_stream(result.iter_content(chunk_size=block_size)), sep=';')
 
         return df
 
@@ -203,6 +233,9 @@ class ODSClient(object):
                           timezone=None,               # type: str
                           use_labels_for_header=True,  # type: bool
                           csv_separator=';',           # type: str
+                          tqdm=False,                  # type: bool
+                          to_path=None,                # type: Union[str, Path]
+                          block_size=1024,             # type: int
                           **other_opts
                           ):
         """
@@ -213,6 +246,9 @@ class ODSClient(object):
         :param timezone:
         :param use_labels_for_header:
         :param csv_separator: ';', ','...
+        :param tqdm: a boolean indicating if a progress bar using tqdm should be displayed. tqdm should be installed
+        :param to_path: a string indicating the file path where to write the csv. In that case nothing is returned
+        :param block_size: an int block size used in streaming mode when to_csv or tqdm is used
         :param other_opts:
         :return:
         """
@@ -247,10 +283,97 @@ class ODSClient(object):
         url = self.get_download_url(dataset_id)
 
         # Execute call
-        result = self._http_call(url, params=opts)
+        result = None
+        if not tqdm:
+            if to_path is None:
+                # return csv string
+                result = self._http_call(url, params=opts)
+            else:
+                # stream to csv file
+                r = self._http_call(url, params=opts, stream=True)
+                with open(str(to_path), 'wb') as f:
+                    for data in r.iter_content(block_size):
+                        f.write(data)
+        else:
+            # we need streaming mode anyway
+            r = self._http_call(url, params=opts, stream=True)
+            total_size = int(r.headers.get('Content-Length', 0))
+
+            from tqdm import tqdm as _tqdm
+            with _tqdm(desc=url, total=total_size,
+                       unit='B' if block_size == 1024 else 'it',
+                       unit_scale=True,
+                       unit_divisor=block_size
+                       ) as bar:
+                if to_path is None:
+                    # stream to a string in memory
+                    result = io.StringIO()
+                    for data in r.iter_content(block_size):
+                        bar.update(len(data))
+                        result.write(data.decode(r.encoding))
+                    result = result.getvalue()
+                else:
+                    # stream to csv file: directly transfer the bytes
+                    with open(str(to_path), 'wb') as f:
+                        for data in r.iter_content(block_size):
+                            bar.update(len(data))
+                            f.write(data)
+            if total_size != 0 and t.n != total_size:
+                print("ERROR, something went wrong")
 
         return result
 
+    # noinspection PyShadowingBuiltins
+    def push_dataset_realtime(self,
+                              dataset_id,         # type: str
+                              dataset,            # type: Union[str, pandas.DataFrame]
+                              push_key,           # type: str
+                              format='csv',       # type: str
+                              csv_separator=';',  # type: str
+                              **other_opts
+                              ):
+        """
+        Pushes a Dataset. This functions accepts either a Pandas Dataframe or a CSV string with header included.
+
+        :param dataset_id:
+        :param dataset: The dataset to push as a list of dicts, where the dict keys are the column names
+        :param push_key: The Push Key provided by the API for pushing this dataset. Warning: This key is independent
+                         from the API key. It can be acquired from the Realtime Push API URL section in ODS.
+        :param format: The format of the dataset to be pushed. Can be `pandas` or `csv`.
+        :param csv_separator: CSV separator character in case of a csv dataset input.
+        :returns: HTTP Response status
+        """
+
+        if format == 'pandas':
+            try:
+                import pandas as pd
+            except ImportError as e:
+                raise Exception("`push_dataset_realtime` with the `pandas` format requires `pandas` to be installed. [%s] %s" % (e.__class__, e))
+            # noinspection PyStatementEffect
+            dataset  # type:pandas.DataFrame
+            request_body = dataset.to_json(orient='records')
+        elif format == 'csv':
+            try:
+                import csv
+            except ImportError as e:
+                raise Exception("`push_dataset_realtime` with the `csv` format requires `csv` to be installed. [%s] %s" % (e.__class__, e))
+            # noinspection PyStatementEffect
+            dataset  # type:str
+            csv_reader = csv.DictReader(StringIO(dataset), delimiter=csv_separator)
+            request_body = dumps([r for r in csv_reader])
+        else:
+            raise ValueError("Dataset format must be either `pandas` or `csv`")
+
+        # Combine all the options
+        opts = other_opts
+        opts['pushkey'] = push_key
+
+        # The URL to call
+        url = self.get_realtime_push_url(dataset_id)
+
+        # Execute call
+        return self._http_call(url, method='post', body=request_body, params=opts, decode=False)
+
     def store_apikey_in_keyring(self,
                                 apikey=None  # type: str
                                 ):
@@ -324,6 +447,7 @@ class ODSClient(object):
                 while k.endswith('/'):
                     k = k[:-1]
                 return k
+
             apikeys_dct = {_remove_trailing_slash(k): v for k, v in apikeys_dct.items()}
 
             # Try to get a match in the dict: first platform id, then base url, then default
@@ -377,6 +501,17 @@ class ODSClient(object):
         """
         return "%s/explore/dataset/%s/download/" % (self.base_url, dataset_id)
 
+    def get_realtime_push_url(self,
+                              dataset_id,  # type: str
+                              ):
+        # type: (...) -> str
+        """
+
+        :param dataset_id:
+        :return:
+        """
+        return "%s/api/push/1.0/%s/realtime/push/" % (self.base_url, dataset_id)
+
     def _http_call(self,
                    url,           # type: str
                    body=None,     # type: bytes
@@ -443,6 +578,7 @@ class NoODSAPIKeyFoundError(Exception):
     Raised when no api key was found (no explicit api key provided, no api key file, no env variable entry, no keyring
     entry)
     """
+
     def __init__(self,
                  odsclient  # type: ODSClient
                  ):
@@ -463,6 +599,7 @@ class InsufficientRightsForODSResourceError(Exception):
     Raised when a HTTP 200 is received from ODS together with an HTML page as body. This happens when api key is
     missing or does not grant the appropriate rights for the required resource.
     """
+
     def __init__(self, headers, contents):
         self.headers = headers
         self.contents = contents
@@ -477,6 +614,7 @@ class ODSException(Exception):
     """
     An error returned by the ODS API
     """
+
     def __init__(self, status_code, headers, **details):
         """
 
@@ -537,7 +675,7 @@ def create_session_for_proxy(http_proxyhost,                  # type: str
     s = Session()
     s.proxies = {
         'http': 'http://%s:%s' % (http_proxyhost, http_proxyport),
-        'https': '%s://%s:%s' % (https_proxy_protocol, https_proxyhost,  https_proxyport),
+        'https': '%s://%s:%s' % (https_proxy_protocol, https_proxyhost, https_proxyport),
     }
     if not (ssl_verify is None):
         s.verify = ssl_verify
@@ -548,7 +686,7 @@ def create_session_for_proxy(http_proxyhost,                  # type: str
     return s
 
 
-def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
+def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE, progressbar=None):
     """
     Lets you use an iterable (e.g. a generator) that yields bytestrings as a read-only
     input stream.
@@ -558,6 +696,7 @@ def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
 
     Source: https://stackoverflow.com/a/20260030/7262247
     """
+
     class IterStream(io.RawIOBase):
         def __init__(self):
             self.leftover = None
@@ -571,7 +710,10 @@ def iterable_to_stream(iterable, buffer_size=io.DEFAULT_BUFFER_SIZE):
                 chunk = self.leftover or next(iterable)
                 output, self.leftover = chunk[:ln], chunk[ln:]
                 b[:len(output)] = output
+                if progressbar:
+                    progressbar.update(len(output))
                 return len(output)
             except StopIteration:
-                return 0    # indicate EOF
+                return 0  # indicate EOF
+
     return io.BufferedReader(IterStream(), buffer_size=buffer_size)
